package ru.dbotthepony.mc.otm.graph.matter

import com.google.common.collect.Streams
import net.minecraft.server.level.ServerLevel
import net.minecraft.world.item.Item
import net.minecraft.world.level.block.entity.BlockEntity
import ru.dbotthepony.mc.otm.capability.FlowDirection
import ru.dbotthepony.mc.otm.capability.MatteryCapability
import ru.dbotthepony.mc.otm.capability.matter.*
import ru.dbotthepony.mc.otm.core.filterNotNull
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.graph.Abstract6Graph
import ru.dbotthepony.mc.otm.graph.Graph6Node
import java.util.*
import java.util.function.Predicate
import java.util.stream.Stream
import kotlin.collections.HashSet

@Suppress("unused")
class MatterNetworkGraph : Abstract6Graph<IMatterGraphNode>(), IMatterGraphListener {
	private val listeners = HashSet<IMatterGraphListener>()

	fun addListener(listener: IMatterGraphListener) = listeners.add(listener)
	fun removeListener(listener: IMatterGraphListener) = listeners.remove(listener)

	override fun onNodeRemoved(node: Graph6Node<IMatterGraphNode>) {
		val patterns = node.value.getPatternHandler()

		if (patterns != null) {
			for (pattern in patterns.patterns) {
				onPatternRemoved(pattern)
			}
		}

		val tasks = node.value.getTaskHandler()

		if (tasks != null) {
			for (task in tasks.replicationTasks) {
				onMatterTaskRemoved(task)
			}
		}

		for (pattern in this.patterns) {
			node.value.onPatternRemoved(pattern)
		}

		for (task in this.tasks) {
			node.value.onMatterTaskRemoved(task)
		}
	}

	override fun onNodeAdded(node: Graph6Node<IMatterGraphNode>) {
		for (pattern in this.patterns) {
			node.value.onPatternAdded(pattern)
		}

		for (task in this.tasks) {
			node.value.onMatterTaskCreated(task)
		}

		val patterns = node.value.getPatternHandler()

		if (patterns != null) {
			for (pattern in patterns.patterns) {
				onPatternAdded(pattern)
			}
		}

		val tasks = node.value.getTaskHandler()

		if (tasks != null) {
			for (task in tasks.replicationTasks) {
				onMatterTaskCreated(task)
			}
		}
	}

	fun getMatterStorageLevel(): Decimal {
		var level = Decimal.ZERO

		for (node in nodes) {
			val matter = node.value.getMatterHandler()

			if (matter != null && matter.matterFlow == FlowDirection.BI_DIRECTIONAL) {
				level += matter.storedMatter
			}
		}

		return level
	}

	fun getMatterStorageMaxLevel(): Decimal {
		var level = Decimal.ZERO

		for (node in nodes) {
			val matter = node.value.getMatterHandler()

			if (matter != null && matter.matterFlow == FlowDirection.BI_DIRECTIONAL) {
				level += matter.maxStoredMatter
			}
		}

		return level
	}

	fun extractMatter(howMuch: Decimal, simulate: Boolean): Decimal {
		if (howMuch <= Decimal.ZERO)
			return Decimal.ZERO

		@Suppress("name_shadowing")
		var howMuch = howMuch
		var extracted = Decimal.ZERO

		for (node in nodes) {
			val matter = node.value.getMatterHandler()

			if (matter != null) {
				val value = matter.extractMatter(howMuch, simulate)
				howMuch -= value
				extracted += value

				if (howMuch <= Decimal.ZERO)
					return extracted
			}
		}

		return extracted
	}

	fun receiveMatter(howMuch: Decimal, simulate: Boolean): Decimal {
		if (howMuch <= Decimal.ZERO)
			return Decimal.ZERO

		@Suppress("name_shadowing")
		var howMuch = howMuch
		var received = Decimal.ZERO

		for (node in nodes) {
			val matter = node.value.getMatterHandler()

			if (matter != null && matter.matterFlow == FlowDirection.BI_DIRECTIONAL) {
				val value = matter.receiveMatter(howMuch, simulate)
				howMuch -= value
				received += value

				if (howMuch <= Decimal.ZERO)
					return received
			}
		}

		return received
	}

	fun receiveMatterForce(howMuch: Decimal, simulate: Boolean): Decimal {
		if (howMuch <= Decimal.ZERO)
			return Decimal.ZERO

		@Suppress("name_shadowing")
		var howMuch = howMuch
		var received = Decimal.ZERO

		for (node in nodes) {
			val matter = node.value.getMatterHandler()

			if (matter != null && matter.matterFlow != FlowDirection.OUTPUT) {
				val value = matter.receiveMatter(howMuch, simulate)
				howMuch -= value
				received += value

				if (howMuch <= Decimal.ZERO)
					return received
			}
		}

		return received
	}

	private fun doInsertPattern(state: IPatternState, onlyUpdate: Boolean, simulate: Boolean): PatternInsertStatus {
		for (node in nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				val status = storage.insertPattern(state, onlyUpdate, simulate)

				if (!status.isFailed) {
					return status
				}
			}
		}

		return PatternInsertFailure
	}

	fun insertPattern(state: IPatternState, onlyUpdate: Boolean, simulate: Boolean): PatternInsertStatus {
		if (onlyUpdate) return doInsertPattern(state, true, simulate)
		val status = doInsertPattern(state, true, simulate)
		if (!status.isFailed) return status
		return doInsertPattern(state, false, simulate)
	}

	val tasks: Stream<out IReplicationTask<*>> get() {
		return Streams.concat(*nodes.mapNotNull { it.value.getTaskHandler()?.replicationTasks }.toTypedArray())
	}

	val allTasks: Stream<out IReplicationTask<*>> get() {
		return Streams.concat(*nodes.mapNotNull { it.value.getTaskHandler()?.allReplicationTasks }.toTypedArray())
	}

	val patterns: Stream<out IPatternState> get() {
		return nodes.stream().map { it.value.getPatternHandler()?.patterns }.filterNotNull().flatMap { it }
	}

	val patternCount: Long get() {
		var value = 0L

		for (node in nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				value += storage.storedPatterns
			}
		}

		return value
	}

	val patternCapacity: Long get() {
		var value = 0L

		for (node in nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				value += storage.patternCapacity
			}
		}

		return value
	}

	fun getPattern(id: UUID): IPatternState? {
		for (node in nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				val get = storage.getPattern(id)

				if (get != null) {
					return get
				}
			}
		}

		return null
	}

	fun getPattern(state: PatternState) = getPattern(state.id)
	fun hasPattern(state: PatternState) = getPattern(state.id) != null
	fun hasPattern(id: UUID) = getPattern(id) != null

	fun findPattern(predicate: Predicate<IPatternState>): IPatternState? {
		for (node in nodes) {
			val storage = node.value.getPatternHandler()

			if (storage != null) {
				val find = storage.patterns.filter(predicate).findAny()

				if (find.isPresent) {
					return find.get()
				}
			}
		}

		return null
	}

	fun findPattern(predicate: Item) = findPattern { it.item == predicate }

	fun findPatterns(predicate: Predicate<IPatternState>): List<IPatternState> {
		return patterns.filter(predicate).toList()
	}

	fun findPatterns(predicate: Item) = findPatterns { it.item == predicate }

	fun allocateTask(simulate: Boolean): ReplicationTaskAllocation? {
		for (node in nodes) {
			val tasks = node.value.getTaskHandler()

			if (tasks != null) {
				val allocated = tasks.allocateTask(simulate)

				if (allocated != null) {
					return allocated
				}
			}
		}

		return null
	}

	fun notifyTaskCompletion(taskId: UUID): Boolean {
		return nodes.any { it.value.getTaskHandler()?.notifyTaskCompletion(taskId) == true }
	}

	override fun onPatternAdded(state: IPatternState) {
		for (node in nodes) node.value.onPatternAdded(state)
		for (node in listeners) node.onPatternAdded(state)
	}

	override fun onPatternRemoved(state: IPatternState) {
		for (node in nodes) node.value.onPatternRemoved(state)
		for (node in listeners) node.onPatternRemoved(state)
	}

	override fun onPatternUpdated(newState: IPatternState, oldState: IPatternState) {
		for (node in nodes) node.value.onPatternUpdated(newState, oldState)
		for (node in listeners) node.onPatternUpdated(newState, oldState)
	}

	override fun onMatterTaskCreated(task: IReplicationTask<*>) {
		for (node in nodes) node.value.onMatterTaskCreated(task)
		for (node in listeners) node.onMatterTaskCreated(task)
	}

	override fun <T : IReplicationTask<*>> onMatterTaskUpdated(newState: T, oldState: T) {
		for (node in nodes) node.value.onMatterTaskUpdated(newState, oldState)
		for (node in listeners) node.onMatterTaskUpdated(newState, oldState)
	}

	override fun onMatterTaskFinished(state: IReplicationTask<*>) {
		for (node in nodes) node.value.onMatterTaskFinished(state)
		for (node in listeners) node.onMatterTaskFinished(state)
	}

	override fun onMatterTaskRemoved(state: IReplicationTask<*>) {
		for (node in nodes) node.value.onMatterTaskRemoved(state)
		for (node in listeners) node.onMatterTaskRemoved(state)
	}

	companion object {
		@JvmStatic
		fun discoverFull(tile: BlockEntity, node: Graph6Node<IMatterGraphNode>) {
			if (tile.level !is ServerLevel)
				return

			return discoverFull(
				tile.level!! as ServerLevel,
				tile.blockPos,
				node,
				fun(_tile): Graph6Node<IMatterGraphNode>? {
					val resolve = _tile.getCapability(MatteryCapability.MATTER_NODE)

					return if (resolve.isPresent) {
						resolve.resolve().get().matterNode
					} else {
						null
					}
				},
				::MatterNetworkGraph
			)
		}

		@JvmStatic
		fun discover(tile: BlockEntity, node: Graph6Node<IMatterGraphNode>): Boolean {
			if (tile.level !is ServerLevel)
				return false

			return discover(
				tile.level!! as ServerLevel,
				tile.blockPos,
				node,
				fun(_tile): Graph6Node<IMatterGraphNode>? {
					val resolve = _tile.getCapability(MatteryCapability.MATTER_NODE)

					return if (resolve.isPresent) {
						resolve.resolve().get().matterNode
					} else {
						null
					}
				},
				::MatterNetworkGraph
			)
		}
	}
}
